/*
    Fichier trtentete.c
    Auteur Bernard Chardonneau

    Logiciel libre, droits d'utilisation prciss en franais
    dans le fichier : licence-fr.txt

    Traductions des droits d'utilisation dans les fichiers :
    licence-de.txt , licence-en.txt , licence-es.txt ,
    licence-it.txt , licence-nl.txt , licence-pt.txt ,
    licence-eo.txt , licence-eo-utf.txt

    Droits d'utilisation galement sur la page web :
    http://libremail.tuxfamily.org/voir.php?page=droits


    Bibliothque de fonctions permettant de

    - reprer les champs de l'entte du mail
    - convertir dans ces champs les accents et caractres spciaux
      cods sur 7 bits grace  l'encodage quoted-printable
    - mmoriser les lignes de l'entte aprs conversion
*/

#include <stdio.h>
#include "buflect.h"
#include "encodage.h"
#include "trtentete.h"


/* cherche si la dernire ligne lue commence par un mot cl particulier */

int start (char *motcle)
{
    int i;

    i = 0;

    // on teste caractre par caractre en ignorant la casse
    while (motcle [i] && tolower (buf_lect [i]) == tolower (motcle [i]))
        i++;

    // le mot cl doit tre suivi de  :  dans la ligne lue
    return ((buf_lect [i] == ':') && (motcle [i] == '\0'));
}



/* convertit la ligne d'entte en interprtant les caractres spciaux */

void majlignentete ()
{
    int encodage;    // mode d'encodage
    int charset;     // jeu de caractres
    int carcode;     // caractre correspondant  un codage quoted-printable
    int decalsuiv;   // indicateur pour conversion de caractre encod Utf8
    int decode_;     // indicateur pour conversion '_' en ' '
    char carcourant; // caractre contenant une valeur en base64
    int valcar [4];  // valeur aprs conversion des caractres encods base64
    int i, j, k, l;  // compteurs


    // initialisation
    i = 0;
    j = 0;
    encodage = defaut;
    charset  = defaut;
    decode_  = 1;

    // exploration de toute la ligne
    while (buf_lect [i] != '\0')
    {
        // si dbut de mot d'encodage
        if (buf_lect [i] == '=' && buf_lect [i+1] == '?' && encodage == defaut)
        {
            // se positionner sur le charset
            i = i + 2;

            // identification simplifie du charset (sur 1 caractre)
            if (tolower (buf_lect [i]) == 'i')
                charset = Iso8859;
            else if (tolower (buf_lect [i]) == 'u')
            {
                charset = Utf8;
                decalsuiv = 0;  // initialisation indicateur de conversion
            }
            else
                charset = xCharset;  // sera trait comme ISO-8859

            // aller  la fin du charset
            do
                i++;
            while (buf_lect [i] && buf_lect [i] != '?');

            // se positionner sur le caractre d'encodage
            i++;

            // identification du mode d'encodage
            if (tolower (buf_lect [i]) == 'q')
                encodage = QuotedPrint;
            else if (tolower (buf_lect [i]) == 'b')
                encodage = Base64;
            else
                encodage = xBits;

            // aller au dbut de la zone encode
            i = i + 2;
        }

        // si fin de zone encode
        if (buf_lect [i] == '?' && buf_lect [i+1] == '=' && encodage != defaut)
        {
            // sauter le marqueur de fin
            i = i + 2;

            // rinitialiser les indicateurs d'tat
            charset  = defaut;
            encodage = defaut;
        }

        // sinon si encodage quoted printable en cours
        else if (buf_lect [i] == '=' && encodage == QuotedPrint
                 && ((buf_lect [i+1] >= '0' && buf_lect [i+1] <= '9')
                  || (buf_lect [i+1] >= 'A' && buf_lect [i+1] <= 'F')))
        {
            // on rcupre le caractre correspondant
            carcode = hexa (buf_lect + ++i);

            // si jeu de caractres ISO-8859 dans la ligne d'entte
            if (charset != Utf8)
            {
                // si le caractre est un caractre de controle
                if (carcode < 0x20)
                {
                    // et que c'est un caractre autre que CR
                    if (carcode != '\r')
                        // le remplacer par un blanc pour
                        // limiter les problmes d'affichage
                        buf_lect [j++] = ' ';
                }

                // sinon si c'est un caractre accentu
                // et que l'on fait un affichage UTF-8
                else if ((carcode & 0x80) && util_utf8 ())
                {
                    // si caractre entre 0xC0 et 0xFF
                    if (carcode & 0x40)
                    {
                        // mmoriser le caractre converti en UTF-8
                        buf_lect [j++] = 0xC3;
                        buf_lect [j++] = carcode - 0x40;
                    }
                    // sinon (caractre entre 0x80 et 0xFF)
                    else
                    {
                        // mmoriser le caractre converti en UTF-8
                        buf_lect [j++] = 0xC2;
                        buf_lect [j++] = carcode;
                    }
                }
                // sinon (cas gnral)
                else
                    // mmoriser le caractre sans conversion
                    buf_lect [j++] = carcode;
            }
            // sinon, jeu de caractres UTF-8
            else
            {
                // si on affiche les caractres en UTF-8
                if (util_utf8 ())
                    // mmoriser le caractre sans conversion
                    buf_lect [j++] = carcode;

                // sinon si caractre  on convertira le caractre qui suit
                else if (carcode == 0xC3)
                    decalsuiv = 1;

                // on recopie les caractres autres que  et CR
                else if (carcode != 0xC2 && carcode != '\r')
                {
                    if (decalsuiv)
                    {
                        // en faisant une conversion si ncessaire
                        carcode = carcode | 0x40;
                        decalsuiv = 0;
                    }

                    buf_lect [j++] = carcode;
                }
            }

            // se positionner sur le caractre suivant
            i = i + 2;
        }

        // sinon si encodage base64 en cours
        else if (encodage == Base64)
        {
            // dcoder une srie de 4 caractres
            for (k = 0; k < 4; k++)
            {
                // rcuprer un caractre
                carcourant = buf_lect [i++];

                // chercher sa valeur sur 6 bits et la mmoriser
                if ('A' <= carcourant && carcourant <= 'Z')
                    valcar [k] = carcourant - 'A';
                else if ('a' <= carcourant && carcourant <= 'z')
                    valcar [k] = carcourant + 26 - 'a';
                else if ('0' <= carcourant && carcourant <= '9')
                    valcar [k] = carcourant + 52 - '0';
                else if (carcourant == '+')
                    valcar [k] = 62;
                else if (carcourant == '/')
                    valcar [k] = 63;
            }

            // reconstituer les 3 caractres dcods correspondants
            k = 0;
            valcar [k++] = (valcar [0] << 2) | (valcar [1] >> 4);

            // s'il y avait un = , a faisait moins de 3 caractres  obtenir
            if (buf_lect [i-2] != '=')
            {
                valcar [k++] = (valcar [1] << 4) | (valcar [2] >> 2);

                if (carcourant != '=')
                    valcar [k++] = (valcar [2] << 6) | (valcar [3]);
            }

            // si jeu de caractres ISO-8859 et on fait un affichage UTF-8
            if (charset != Utf8 && util_utf8 ())
            {
                // convertir caractres de ISO-8859 en UTF-8 et les mmoriser
                // le test j < i permet d'viter d'craser la chaine initiale
                for (l= 0; l < k && j < i; l++)
                {
                    // si le caractre est un caractre de controle
                    if (valcar [l] < 0x20)
                    {
                        // et que c'est un caractre autre que CR
                        if (valcar [l] != '\r')
                            // le remplacer par un blanc pour
                            // limiter les problmes d'affichage
                            buf_lect [j++] = ' ';
                    }
                    // sinon si caractre spcial
                    else if (valcar [l] & 0x80)
                    {
                        // si caractre entre 0xC0 et 0xFF
                        if (valcar [l] & 0x40)
                        {
                            // mmoriser le caractre converti en UTF-8
                            buf_lect [j++] = 0xC3;
                            buf_lect [j++] = valcar [l] - 0x40;
                        }
                        // sinon (caractre entre 0x80 et 0xFF)
                        else
                        {
                            // mmoriser le caractre converti en UTF-8
                            buf_lect [j++] = 0xC2;
                            buf_lect [j++] = valcar [l];
                        }
                    }
                    // sinon (cas gnral)
                    else
                        // mmoriser le caractre sans conversion
                        buf_lect [j++] = valcar [l];
                }
            }
            // sinon si jeu de caractres UTF-8 et on fait un affichage ISO-8859
            else if (charset == Utf8 && (! util_utf8 ()))
            {
                // convertir caractres de UTF-8 en ISO-8859 et les mmoriser
                for (l= 0; l < k; l++)
                {
                    // si le caractre est un caractre de controle
                    if (valcar [l] < 0x20)
                    {
                        // et que c'est un caractre autre que CR
                        if (valcar [l] != '\r')
                            // le remplacer par un blanc pour
                            // limiter les problmes d'affichage
                            buf_lect [j++] = ' ';
                    }
                    // sinon si caractre spcial
                    else if (valcar [l] & 0x80)
                    {
                        // si le caractre prcdent tait 
                        if (buf_lect [j-1] == 0xC2)
                            // on le remplace par le caractre courant
                            buf_lect [j-1] = valcar [l];

                        // sinon si le caractre prcdent tait 
                        else if (buf_lect [j-1] == 0xC3)
                            // on le remplace par le caractre courant converti
                            buf_lect [j-1] = valcar [l] | 0x40;
                        else
                            // sinon mmoriser le caractre sans conversion
                            buf_lect [j++] = valcar [l];
                    }
                    // sinon (cas gnral)
                    else
                        // mmoriser le caractre sans conversion
                        buf_lect [j++] = valcar [l];
                }
            }
            // sinon (jeu de caractre lu = jeu de caractres affich)
            else
            {
                // simple mmorisation des caractres dcods
                for (l= 0; l < k; l++)
                {
                    // si le caractre est un caractre de controle
                    if (valcar [l] < 0x20)
                    {
                        // et que c'est un caractre autre que CR
                        if (valcar [l] != '\r')
                            // le remplacer par un blanc pour
                            // limiter les problmes d'affichage
                            buf_lect [j++] = ' ';
                    }
                    // sinon (cas gnral)
                    else
                        // mmoriser le caractre sans conversion
                        buf_lect [j++] = valcar [l];
                }
            }
        }

        // sinon, si on a le caractre spcial _
        else if (buf_lect [i] == '_' && decode_)
        {
            // le remplacer par un blanc
            buf_lect [j++] = ' ';

            // et passer au caractre suivant
            i++;
        }

        // sinon, si on a un caractre de controle
        else if (buf_lect [i] < 0x20)
        {
            // et que c'est un caractre autre que CR
            if (buf_lect [i++] != '\r')
                // le remplacer par un blanc pour viter problmes d'affichage
                buf_lect [j++] = ' ';
        }
        // sinon (cas gnral : ni encodage ni caractre spcial)
        else
        {
            // prendre en compte les < et > pour viter de
            // convertir les _ en ' ' dans une adresse email
            if (buf_lect [i] == '<')
                decode_ = 0;
            else if (buf_lect [i] == '>')
                decode_ = 1;

            // on conserve le caractre courant et on passe au suivant
            // remarque : on n'a pas jug utile de convertir le jeu de
            //            caractres accentus vu qu'il ne devrait pas
            //            y en avoir sans encodage
            buf_lect [j++] = buf_lect [i++];
        }
    }

    // terminaison de la chaine de caractres
    buf_lect [j] = '\0';
}



/* convertit la ligne lue en interprtant les caractres
   spciaux et la mmorise dans le buffer de destination */

void memconvbuf (char *bufdest)
{
    // on convertit les caractres spciaux
    majlignentete ();

    // mmorisation de la ligne aprs conversion
    membuf (bufdest);
}



/* mmorise la ligne lue en dans le buffer de destination */

void membuf (char *bufdest)
{
    int i;

    i = 0;

    // copie tronque  la longueur max
    do
        bufdest [i] = buf_lect [i++];
    while (i < 79 && buf_lect [i] != '\0' && buf_lect [i] != '\n');

    // terminaison de la chaine de caractres
    bufdest [i] = '\0';
}



/* retourne le code hexa lu  la position courante */

int hexa (char *chaine)
{
    int valcar;
    char souschaine [3];

    souschaine [0] = chaine [0];
    souschaine [1] = chaine [1];
    souschaine [2] = '\0';

    sscanf (souschaine, "%x", &valcar);
    return valcar;
}
